#ifndef MYX_CORE_PATHS_CPP_
#define MYX_CORE_PATHS_CPP_

#include <myx/core/config.hpp>
#include <myx/filesystem/paths.hpp>

#include <paths.h>

#include <QCoreApplication>
#include <QString>

#ifndef QT_NO_DEBUG
#include <QDebug>
#endif

namespace myx {

namespace filesystem {

MYXLIB_INLINE Paths::Paths() :
	m_configFileExtension( ".conf" ),
	m_binDirRegex        ( "/s*bin$" ),
	m_unityBinDirRegex   ( "/bin/unity$" )
{
	QFileInfo procSelfExe( QStringLiteral( "/proc/self/exe" ) );
	QFileInfo currentExecutable = procSelfExe.canonicalFilePath();
	m_executableName      = currentExecutable.fileName();
	m_executableFilePath  = currentExecutable.absoluteFilePath();
	m_executableDirectory = currentExecutable.absolutePath();
}


MYXLIB_INLINE bool Paths::init()
{
	m_configFileName = m_executableName + m_configFileExtension;
	return( initCommon() );
}


//MYXLIB_INLINE bool Paths::init( const QString& configFileName )
//{
//	m_configFileName = configFileName;
//	return( initCommon() );
//}


//MYXLIB_INLINE bool Paths::init( const QString& projectName, const QString& configFileExtension )
//{
//	m_projectName = projectName.isEmpty() ? m_executableName
//	                                      : projectName;
//	auto ext = configFileExtension.isEmpty() ? QStringLiteral( "conf" )
//	                                         : configFileExtension;
//	m_configFileName = m_projectName + "." + ext;
//	return( initCommon() );
//}


MYXLIB_INLINE bool Paths::initCommon()
{
	m_homeDirectory = QDir::homePath();
	m_tempDirectory = QDir::tempPath();

	m_configDirectory = QString::fromLocal8Bit( qgetenv( "XDG_CONFIG_HOME" ) );
	if ( m_configDirectory.isEmpty() )
	{
		m_configDirectory = m_homeDirectory + "/.config";
	}

	m_dataDirectory = QString::fromLocal8Bit( qgetenv( "XDG_DATA_HOME" ) );
	if ( m_dataDirectory.isEmpty() )
	{
		m_dataDirectory = m_homeDirectory + "/.local/share";
	}

	m_hierarchyType = getHierarchyType();
	calculatePaths( m_hierarchyType );

	m_configFilePath = m_systemConfigDirectory + "/" + m_configFileName;

	return( true );
} // Paths::initCommon


MYXLIB_INLINE void Paths::setupSystemDirectories( const QString& defaultProjectDirectory,
                                                  const QString& defaultEtcDirectory,
                                                  const QString& defaultConstDataDirectory,
                                                  const QString& defaultVarDataDirectory,
                                                  const QString& defaultLogDirectory )
{
	QFileInfo prefixDirInfo { defaultProjectDirectory };
	if ( prefixDirInfo.isDir() && prefixDirInfo.isReadable() )
	{
		m_projectDirectory = defaultProjectDirectory;
	}
	else
	{
		m_projectDirectory = QStringLiteral( "." );
	}

	QFileInfo etcDirInfo { defaultEtcDirectory };
	if ( etcDirInfo.isDir() && etcDirInfo.isReadable() )
	{
		m_systemConfigDirectory = defaultEtcDirectory;
	}
	else
	{
		m_systemConfigDirectory = m_userConfigDirectory;
	}

	QFileInfo constDataDirInfo { defaultConstDataDirectory };
	if ( constDataDirInfo.isDir() && constDataDirInfo.isReadable() )
	{
		m_systemConstDataDirectory = defaultConstDataDirectory;
	}
	else
	{
		m_systemConstDataDirectory = m_userConstDataDirectory;
	}

	QFileInfo varDataDirInfo { defaultVarDataDirectory };
	if ( varDataDirInfo.isDir() && varDataDirInfo.isWritable() )
	{
		m_systemVarDataDirectory = defaultVarDataDirectory;
	}
	else
	{
		m_systemVarDataDirectory = m_userVarDataDirectory;
	}

	QFileInfo logDirInfo { defaultLogDirectory };
	if ( logDirInfo.isDir() && logDirInfo.isWritable() )
	{
		m_systemLogDirectory = defaultLogDirectory;
	}
	else
	{
		m_systemLogDirectory = m_userLogDirectory;
	}
} // Paths::setupSystemDirectories


MYXLIB_INLINE void Paths::setupUserDirectories( const QString& prefix )
{
	m_userConfigDirectory    = m_configDirectory + "/" + prefix;
	m_userConstDataDirectory = m_dataDirectory + "/" + prefix + "/share";
	m_userVarDataDirectory   = m_dataDirectory + "/" + prefix + "/var";
	m_userLogDirectory       = m_dataDirectory + "/" + prefix + "/log";
}


MYXLIB_INLINE Paths::HierarchyType Paths::getHierarchyType()
{
	if ( m_unityBinDirRegex.match( m_executableDirectory ).hasMatch() )
	{
		return ( HierarchyType::kDevelopment );
	}

	if ( !m_binDirRegex.match( m_executableDirectory ).hasMatch() )
	{
		return ( HierarchyType::kFlat );
	}

	if ( m_executableDirectory.startsWith( QStringLiteral( "/opt/" ) ) )
	{
		return ( HierarchyType::kOpt );
	}

	if ( m_executableDirectory.startsWith( QStringLiteral( "/usr/local/bin/" ) ) )
	{
		return ( HierarchyType::kUsrLocal );
	}

	if ( m_executableDirectory.startsWith( QStringLiteral( "/usr/local/" ) ) )
	{
		return ( HierarchyType::kUsrLocalOrg );
	}

	if ( m_executableDirectory.startsWith( QStringLiteral( "/usr/bin/" ) ) )
	{
		return ( HierarchyType::kUsr );
	}

	if ( m_executableDirectory.startsWith( m_homeDirectory + "/.local/bin/" ) ||
	     m_executableDirectory.startsWith( m_homeDirectory + "/bin/" ) )
	{
		return( HierarchyType::kHome );
	}

	return ( HierarchyType::kDevelopment );
} // Paths::getHierarchyType


MYXLIB_INLINE void Paths::calculatePaths( HierarchyType hType )
{
	auto directory = m_executableDirectory;

	switch ( hType )
	{
	case HierarchyType::kFlat:
		setupSystemDirectories( directory, directory, directory, directory, directory );
		return;

	case HierarchyType::kOpt:
		if ( m_organizationName.isEmpty() || m_themeName.isEmpty() )
		{
			QRegularExpression regex( "^/opt/(.+?)-(.+?)/" );
			QRegularExpressionMatch match = regex.match( m_executableDirectory );
			if ( match.hasMatch() )
			{
				m_organizationName = match.captured( 1 );
				m_themeName        = match.captured( 2 );

				QRegularExpression vr( "(.+?)\\.(.+)" );
				QRegularExpressionMatch vm = vr.match( m_themeName );
				if ( vm.hasMatch() )
				{
					m_themeName = vm.captured( 1 );
					m_version   = vm.captured( 2 );
				}
			}
		}

		if ( m_projectName.isEmpty() )
		{
			QRegularExpression regex( "^/opt/.+/(.+?)/" );
			QRegularExpressionMatch match = regex.match( m_executableDirectory );
			if ( match.hasMatch() )
			{
				m_projectName = match.captured( 1 );
			}
		}
		directory.remove( m_binDirRegex );

		setupSystemDirectories( directory,
		                        directory + "/etc",
		                        directory + "/share",
		                        directory + "/var",
		                        directory + "/log" );
		setupUserDirectories( m_organizationName + "-" + m_themeName + "/" + m_projectName );
		return;

	case HierarchyType::kUsr:
		setupSystemDirectories( QStringLiteral( "/usr" ),
		                        "/etc/" + m_projectName,
		                        "/usr/share/" + m_projectName,
		                        "/var/lib/" + m_projectName,
		                        "/var/log/" + m_projectName );
		return;

	case HierarchyType::kUsrLocal:
		setupSystemDirectories( QStringLiteral( "/usr/local" ),
		                        "/usr/local/etc/" + m_projectName,
		                        "/usr/local/share/" + m_projectName,
		                        "/var/lib/" + m_projectName,
		                        "/var/log/" + m_projectName );
		return;

	case HierarchyType::kUsrLocalOrg:
		directory.remove( m_binDirRegex );
		setupSystemDirectories( directory,
		                        directory + "/etc",
		                        directory + "/share",
		                        directory + "/var",
		                        directory + "/log" );
		return;

	case HierarchyType::kHome:
		m_projectDirectory         = m_homeDirectory;
		m_systemConfigDirectory    = m_userConfigDirectory;
		m_systemConstDataDirectory = m_userConstDataDirectory;
		m_systemVarDataDirectory   = m_userVarDataDirectory;
		m_systemLogDirectory       = m_userLogDirectory;
		return;

	case HierarchyType::kDevelopment:
		directory.remove( m_unityBinDirRegex );
		directory.remove( m_binDirRegex );
		setupSystemDirectories( directory,
		                        directory + "/etc",
		                        directory + "/share",
		                        directory + "/var",
		                        directory + "/log" );
		return;

	case HierarchyType::kUndefined:
		;
	} // switch
} // Paths::calculatePaths


MYXLIB_INLINE bool Paths::makeDefaultSystemDirectories()
{
	bool status = true;
	QDir dir;

	if ( !dir.mkpath( m_systemConfigDirectory ) ) { status = false; }
	if ( !dir.mkpath( m_systemVarDataDirectory ) ) { status = false; }
	if ( !dir.mkpath( m_systemConstDataDirectory ) ) { status = false; }
	if ( !dir.mkpath( m_systemLogDirectory ) ) { status = false; }

	return( status );
}


MYXLIB_INLINE bool Paths::makeDefaultUserDirectories()
{
	bool status = true;
	QDir dir;

	if ( !dir.mkpath( m_userConfigDirectory ) ) { status = false; }
	if ( !dir.mkpath( m_userVarDataDirectory ) ) { status = false; }
	if ( !dir.mkpath( m_userConstDataDirectory ) ) { status = false; }
	if ( !dir.mkpath( m_userLogDirectory ) ) { status = false; }

	return( status );
}


MYXLIB_INLINE bool Paths::makeDefaultDirectories()
{
	return ( makeDefaultSystemDirectories() && makeDefaultUserDirectories() );
}


MYXLIB_INLINE QString Paths::findConfigFile( const QString& defaultConfigFile )
{
	if ( !defaultConfigFile.isEmpty() && QFileInfo( defaultConfigFile ).isReadable() )
	{
		m_configFilePath = defaultConfigFile;
		return( defaultConfigFile );
	}

	auto fileName = QString::fromLocal8Bit( qgetenv( QCoreApplication::applicationName()
	                                                 .toUpper().toUtf8() + "_CONFIG" ) );
	if ( QFileInfo( fileName ).isReadable() )
	{
		m_configFilePath = fileName;
		return( fileName );
	}

	if ( QFileInfo( m_configFilePath ).isReadable() )
	{
		return( m_configFilePath );
	}

	return( QString() );
} // Paths::findConfigFile


MYXLIB_INLINE const QString& Paths::userConfigDirectory() const
{
	return( m_userConfigDirectory );
}


MYXLIB_INLINE const QString& Paths::systemConfigDirectory() const
{
	return( m_systemConfigDirectory );
}


MYXLIB_INLINE const QString& Paths::configFilePath() const
{
	return( m_configFilePath );
}


MYXLIB_INLINE const QString& Paths::configFileName() const
{
	return( m_configFileName );
}


const QString& Paths::configFileExtension() const
{
	return( m_configFileExtension );
}


void Paths::setConfigFileExtension( const QString& name )
{
	m_configFileExtension = name;
}


MYXLIB_INLINE const QString& Paths::userVarDataDirectory() const
{
	return( m_userVarDataDirectory );
}


MYXLIB_INLINE const QString& Paths::systemVarDataDirectory() const
{
	return( m_systemVarDataDirectory );
}


MYXLIB_INLINE const QString& Paths::userConstDataDirectory() const
{
	return( m_userConstDataDirectory );
}


MYXLIB_INLINE const QString& Paths::systemConstDataDirectory() const
{
	return( m_systemConstDataDirectory );
}


MYXLIB_INLINE const QString& Paths::userLogDirectory() const
{
	return( m_userLogDirectory );
}


MYXLIB_INLINE const QString& Paths::systemLogDirectory() const
{
	return( m_systemLogDirectory );
}


MYXLIB_INLINE const QString& Paths::tempDirectory() const
{
	return( m_tempDirectory );
}


MYXLIB_INLINE const QString& Paths::homeDirectory() const
{
	return( m_homeDirectory );
}


MYXLIB_INLINE const QString& Paths::projectName() const
{
	return( m_projectName );
}


MYXLIB_INLINE const QString& Paths::organizationName() const
{
	return( m_organizationName );
}


MYXLIB_INLINE void Paths::setOrganizationName( const QString& name )
{
	m_organizationName = name;
}


MYXLIB_INLINE const QString& Paths::themeName() const
{
	return( m_themeName );
}


MYXLIB_INLINE void Paths::setThemeName( const QString& name )
{
	m_themeName = name;
}


MYXLIB_INLINE const QString& Paths::executableName() const
{
	return( m_executableName );
}


MYXLIB_INLINE const QString& Paths::executableFilePath() const
{
	return( m_executableFilePath );
}


MYXLIB_INLINE const QString& Paths::executableDirectory() const
{
	return( m_executableDirectory );
}


MYXLIB_INLINE const QString& Paths::projectDirectory() const
{
	return( m_projectDirectory );
}

} // namespace filesystem

} // namespace myx

#endif // MYX_CORE_PATHS_CPP_
